# Copyright 2015 Cloudbase Solutions Srl
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

$ErrorActionPreference = 'Stop'

# Charm Helpers
Import-Module JujuLogging
Import-Module JujuWindowsUtils
Import-Module JujuHelper
Import-Module JujuHooks
Import-Module JujuUtils


####################
# GLOBAL VARIABLES #
####################

$DRUPAL_ROOT = Join-Path $env:SystemDrive "inetpub\wwwroot\Drupal"
$DRUPAL_WEBSITE_NAME = "Drupal"
$PHP_LOCATION = Join-Path $env:SystemDrive "PHP"


####################
# MODULE FUNCTIONS #
####################

# Returns the full path to the charm's templates directory
function Get-TemplatesDir {
    return (Join-Path $env:CHARM_DIR "templates")
}

# On Windows Server, IIS is enabled by installing the required Windows
# features for it.
function Install-IIS {
    Write-JujuLog "Installing IIS"

    $iisWindowsFeatures = @( "Web-Server", "Web-Http-Redirect", 
        "Web-DAV-Publishing", "Web-Custom-Logging", "Web-Log-Libraries",
        "Web-ODBC-Logging", "Web-Request-Monitor", "Web-Http-Tracing",
        "Web-Mgmt-Console", "Web-Mgmt-Compat", "Web-Scripting-Tools",
        "Web-Mgmt-Service", "Web-CGI" )
    Install-WindowsFeatures $iisWindowsFeatures

    Write-JujuLog "Finished installing IIS"
}

# For IIS to host PHP applications, a handler mapping must be added that tells
# IIS to pass all PHP-specific requests to the PHP application framework
# by using the FastCGI protocol.
function Start-ConfigureIIS {
    Write-JujuLog "Configuring IIS to handle PHP requests"
    Import-Module WebAdministration

    # Create IIS FastCGI process pool
    $phpCgi = Join-Path $PHP_LOCATION "php-cgi.exe"
    $app = Get-WebConfiguration 'system.webServer/fastCGI/application' | Where-Object { $_.fullPath -eq $phpCgi }
    if (!$app) {
        Add-WebConfiguration 'system.webServer/fastCGI' -Value @{'fullPath' = $phpCgi }
    }

    # Create IIS handler mapping for handling PHP requests
    $phpHandlerName = "PHP_via_FastCGI"
    $phpHandler = Get-WebConfiguration 'system.webServer/handlers/add' | Where-Object { $_.Name -eq $phpHandlerName }
    if (!$phpHandler) {
        Add-WebConfiguration 'system.webServer/handlers' -Value @{
            Name = $phpHandlerName;
            Path = "*.php";
            Verb = "*";
            Modules = "FastCgiModule";
            scriptProcessor=$phpCgi;
            resourceType='Either' 
        }
    }

    # Set PHP environment variables
    $config = Get-WebConfiguration "system.webServer/fastCgi/application[@fullPath='$phpCgi']/environmentVariables/environmentVariable"
    $maxRequests = Get-JujuCharmConfig -Scope "php-max-requests"
    if (!$config) {
        $configPath = "system.webServer/fastCgi/application[@fullPath='$phpCgi']/environmentVariables"
        Add-WebConfiguration $configPath -Value @{ 'Name' = 'PHP_FCGI_MAX_REQUESTS'; Value = $maxRequests }
    }
    $configPath = "system.webServer/fastCgi/application[@fullPath='$phpCgi']"
    Set-WebConfigurationProperty $configPath -Name instanceMaxRequests -Value $maxRequests

    # Restart IIS services
    Start-ExternalCommand { iisreset.exe /restart } -ErrorMessage "Failed to restart IIS services"

    Write-JujuLog "Finished configuring IIS."
}

# Installs a package. Supported packages formats are '.exe' and '.msi'.
function Install-Package {
    Param(
        [Parameter(Mandatory=$true)]
        [string]$InstallerPath,
        [array]$ArgumentList
    )

    Write-JujuLog "Installing package '$InstallerPath'"

    $pkgExtension = $InstallerPath.Split('.')[-1]
    if ($pkgExtension -notin @('msi', 'exe')) {
        Throw ("ERROR: Cannot install the package '$InstallerPath'. " +
               "Unsupported installer format.")
    }

    $stat = Start-Process -FilePath $InstallerPath -ArgumentList $ArgumentList -PassThru -Wait
    if ($stat.ExitCode -ne 0) {
        Throw "$InstallerPath failed to install."
    }
    Remove-Item $installerPath

    Write-JujuLog "Finished installing $InstallerPath"
}

# Parses an url as given in the config option and return a Hashtable with
# the keys: 'url', 'hash' and 'hashing_algorithm'.
# Example:
#       http://example.com/archive.zip#md5=43d89a2f6b8a8918ce3eb76227685276
#
#       returns:
#
#       @{'url' = 'http://example.com/archive.zip';
#         'hash' = 'md5';
#         'hashing_algorithm' = '43d89a2f6b8a8918ce3eb76227685276'}
#
function Get-ParsedUrl {
    Param(
        [Parameter(Mandatory=$true)]
        [Uri]$Url
    )

    $fragment = $Url.Fragment.Trim('#')
    if (!$fragment){
        return @{
            'url' = $Url.AbsoluteUri.Split('#')[0];
            'hash' = $null;
            'hashing_algorithm' = $null
        }
    }
    $details = $fragment.Split("=")
    $algorithm = $details[0]
    $hash = $details[1]
    if($algorithm -notin @("SHA1", "SHA256", "SHA384", "SHA512", "MACTripleDES", "MD5", "RIPEMD160")) {
        $hash = $null
        $algorithm = $null
    }
    return @{
        'url' = $Url.AbsoluteUri.Split('#')[0];
        'hash' = $hash;
        'hashing_algorithm' = $algorithm
    }
}

function Start-DownloadFile {
    Param(
        [Parameter(Mandatory=$True)]
        [System.Uri]$Uri,
        [string]$OutFile,
        [switch]$SkipIntegrityCheck=$false
    )

    if(!$OutFile) {
        $OutFile = $Uri.PathAndQuery.Substring($Uri.PathAndQuery.LastIndexOf("/") + 1)
        if(!$OutFile) {
            throw "The ""OutFile"" parameter needs to be specified"
        }
    }

    $webClient = New-Object System.Net.WebClient
    Start-ExecuteWithRetry { $webClient.DownloadFile($Uri, $OutFile) }

    if(!$SkipIntegrityCheck) {
        $fragment = $Uri.Fragment.Trim('#')
        if (!$fragment){
            return
        }
        $details = $fragment.Split("=")
        $algorithm = $details[0]
        $hash = $details[1]
        if($algorithm -in @("SHA1", "SHA256", "SHA384", "SHA512", "MACTripleDES", "MD5", "RIPEMD160")){
            Test-FileIntegrity -File $OutFile -Algorithm $algorithm -ExpectedHash $hash
        } else {
            Throw "Hash algorithm $algorithm not recognized."
        }
    }
}

# Returns the full path of the file after it is downloaded from the Internet.
# The function accepts as parameter the url of the file to be downloaded.
function Get-DownloadedFilePath {
    Param(
        [Parameter(Mandatory=$true)]
        [Uri]$Url
    )

    $packagePath = Join-Path $env:TEMP $Url.PathAndQuery.Substring($Url.PathAndQuery.LastIndexOf("/") + 1)
    if (Test-Path $packagePath) {
        $parsedUrl = Get-ParsedUrl $Url
        try {
            $out = Test-FileIntegrity -File $packagePath `
                                      -Algorithm $parsedUrl['hashing_algorithm'] `
                                      -ExpectedHash $parsedUrl['hash']
            return $packagePath
        } catch {
            Remove-Item -Path $packagePath
        }
    }
    Write-JujuLog "Downloading '$Url'"
    $out = Start-DownloadFile -Uri $Url -OutFile $packagePath
    return $packagePath
}

# Extract an zip archive into a destination folder received as a parameter.
# The switch parameter 'WithFolder' indicates if the files are
# archived into a folder or not. This parameter is added due to the
# inconsistency of the archives (some of them have the files added
# into a folder, and then the folder is archived).
function Install-ZipArchive {
    Param(
        [Parameter(Mandatory=$true)]
        [string]$ArchivePath,
        [Parameter(Mandatory=$true)]
        [string]$Destination,
        [switch]$WithFolder
    )

    if (Test-Path $Destination) {
        Write-JujuLog ("'$Destination' folder already exists. Probably " +
                       "'$ArchivePath' is already installed.")
        Remove-Item $ArchivePath
        return
    }
    Write-JujuLog "Installing '$ArchivePath'"
    New-Item -ItemType Directory $Destination
    try {
        if ($WithFolder) {
            # Extract the folder from the archive in a temporary directory
            # and move the files to the destination folder
            $tempDir = Join-Path $env:TEMP ("drupal-iis-temp-" + (Get-Random))
            if (Test-Path $tempDir) {
                Remove-Item -Recurse -Force $tempDir
            }
            New-Item -ItemType Directory $tempDir
            Expand-ZipArchive $ArchivePath $tempDir
            Move-Item "$tempDir\*\*" $Destination
            Remove-Item -Recurse -Force $tempDir
        } else {
            Expand-ZipArchive $ArchivePath $Destination
        }
        Remove-Item $ArchivePath
    } catch {
        # Do clean up an throw exception further on
        if (Test-Path $Destination) {
            Remove-Item -Recurse -Force $Destination
        }
        if (Test-Path $tmpDir) {
            Remove-Item -Recurse -Force $tempDir
        }
        Throw $_
    }
    Write-JujuLog "Finished installing '$Url'"
}

# Installs all the prerequisites needed to run Drupal with PHP on IIS.
function Install-DrupalPrerequisites {
    Write-JujuLog "Installing Drupal prerequisites"

    $packages = @(
        @{
            'url' = Get-JujuCharmConfig -Scope 'vc9-url';
            'name' = "Microsoft Visual C++ 2008 Redistributable - x86 9.0.30729.17";
            'unattended_args' = @('/q')
        };
        @{
            'url' = Get-JujuCharmConfig -Scope 'ur-url';
            'name' = "IIS URL Rewrite Module 2";
            'unattended_args' = @('/q')
        }
    )
    foreach($pkg in $packages) {
        if (Get-ComponentIsInstalled $pkg['name']) {
            Write-JujuLog ($pkg['name'] + " is already installed")
            continue
        }
        $installerPath = Get-DownloadedFilePath $pkg['url']
        Install-Package $installerPath $pkg['unattended_args']
    }

    $drushRoot = Join-Path $env:SystemDrive "windrush"
    if (!(Test-Path $drushRoot)) {
        $drushUrl = Get-JujuCharmConfig -Scope 'drush-url'
        $zipPath = Get-DownloadedFilePath $drushUrl
        Install-ZipArchive $zipPath $drushRoot -WithFolder
    } else {
        Write-JujuLog "Drush archive already extracted"
    }
    Add-ToUserPath "$drushRoot;$drushRoot\tools\bin"

    # Point Drush to the PHP version used by Drupal
    $drushBat = Join-Path $drushRoot "drush.bat"
    $content = @(
        "@echo off",
        "SET SCRIPT_HOME=%~dp0",
        "SET PHP_PATH=$PHP_LOCATION",
        "SET PATH=%SCRIPT_HOME%tools\bin;%PHP_PATH%;%PATH%",
        '@php.exe "%SCRIPT_HOME%vendor\drush\drush\drush.php" --php="php.exe" %*'
    )
    Set-Content $drushBat $content

    Write-JujuLog "Finished Drupal prerequisites"
}

function Get-MySQLContext {
    $requiredCtx = @{
        "db_host" = $null;
        "password" = $null;
    }
    $ctxt = Get-JujuRelationContext -Relation "mysql-db" -RequiredContext $requiredCtx
    if(!$ctxt.Count) {
        return @{}
    }
    return @{
        'db_host' = $ctxt['db_host'];
        'db_user_password' = $ctxt['password']
    }
}

function Get-PostgreSQLContext {
    $requiredCtx = @{
        "host" = $null;
        "database" = $null;
        "user" = $null;
        "password" = $null;
    }
    $ctxt = Get-JujuRelationContext 'pgsql-db' -RequiredContext $requiredCtx
    if(!$ctxt.Count) {
        return @{}
    }
    return @{
        'db_host' = $ctxt['host'];
        'db_name' = $ctxt['database'];
        'db_user_name' = $ctxt['user'];
        'db_user_password' = $ctxt['password']
    }
}

function Get-MSSQLContext {
    $requiredCtx = @{
        "private-address" = $null;
        "database" = $null;
        "user" = $null;
        "password" = $null;
        "instance" = $null;
    }
    $ctxt = Get-JujuRelationContext -Relation "mssql-db" -RequiredContext $requiredCtx
    if(!$ctxt.Count) {
        return @{}
    }
    return @{
        'db_host' = $ctxt['private-address'];
        'db_name' = $ctxt['database'];
        'db_user_name' = $ctxt['user'];
        'db_user_password' = $ctxt['password'];
        'db_instance_name' = $ctxt['instance']
    }
}

function Get-MemcachedContext {
    $requiredCtx = @{
        "private-address" = $null;
        "port" = $null;
        "udp-port" = $null;
    }
    $ctxt = Get-JujuRelationContext -Relation "cache" -RequiredContext $requiredCtx
    if(!$ctxt.Count) {
        return @{}
    }
    return @{
        "cache_host"     = $ctxt["private-address"];
        "cache_tcp_port" = $ctxt["port"];
        "cache_udp_port" = $ctxt["udp-port"]
    }
}

function Get-ClusterContext {
    $requiredCtx = @{
        "db-connected" = $null;
        "memcache-installed" = $null
    }
    $ctxt = Get-JujuRelationContext -Relation "cluster" -RequiredContext $requiredCtx
    if(!$ctxt.Count) {
        return @{}
    }
    return $ctxt
}

# Checks if the database is connected or not.
function Get-DatabaseIsConnected {
    $connectedDB = Get-CharmState -Key 'DatabaseIsConnected'
    return ($connectedDB -eq 'True')
}

# Returns true if the Memcache module is installed.
function Get-MemcacheIsInstalled {
    $memecacheInstalled = Get-CharmState -Key 'MemcacheIsInstalled'
    return ($memecacheInstalled -eq 'True')
}

# Returns the context of the database relation. Checks if a relation with
# a database is created by verifying the returned context. In case there is
# a database relation, its context with an additional key 'db_type' is returned
# (this needed for database type identification) Otherwise, $null is returned.
function Get-DatabaseContext {
    $ctxt = Get-MySQLContext
    if ($ctxt.Count) {
        $ctxt['db_type'] = 'mysql'
        $ctxt['db_user_name'] = Get-JujuCharmConfig -Scope 'database-user-name'
        $ctxt['db_name'] = Get-JujuCharmConfig -Scope 'database-name'
        return $ctxt
    }
    $ctxt = Get-MSSQLContext
    if ($ctxt.Count) {
        $ctxt['db_type'] = 'sqlsrv'
        $ctxt['db_host'] = $ctxt['db_host'] + "\" + $ctxt['db_instance_name']
        return $ctxt
    }
    $ctxt = Get-PostgreSQLContext
    if ($ctxt.Count) {
        $ctxt['db_type'] = 'pgsql'
        return $ctxt
    }
    return @{}
}

# Accepts two parameters: 'TemplatePath', full path to the template that
# is used to generate the config file contents and 'Configs', HashTable
# with the variables that will populate the template. The config file contents
# are returned.
function Get-ConfigFileContent {
    Param(
        [Parameter(Mandatory=$true)]
        [string]$TemplatePath,
        [Parameter(Mandatory=$true)]
        [HashTable]$Configs
    )

    $content = Get-Content $TemplatePath
    foreach ($variable in $Configs.GetEnumerator()) {
        $regex = "{{\s*" + $variable.Name + "\s*}}"
        $content = $content | ForEach-Object { $_ -replace $regex,$variable.Value }
    }
    return $content
}

# Generate php.ini file using the template from "$env:CHARM_DIR\templates".
# Full path to php.ini is returned if it was successfully generated or already
# exists.
function Write-PHPIniFile {
    $phpIni = Join-Path $PHP_LOCATION "php.ini"
    if (Test-Path $phpIni) {
        Write-JujuLog "php.ini file already created."
        return $phpIni
    }
    Write-JujuLog "Generating php.ini file"
    $templatePath = Join-Path (Get-TemplatesDir) "php.ini"
    $configs = @{
        "error_log"           = (Join-Path $PHP_LOCATION "PHP_errors.log");
        "temp_directory"      = (Join-Path $env:SystemRoot "temp");
        "max_execution_time"  = (Get-JujuCharmConfig -scope "php-max-execution-time");
        "extensions_location" = (Join-Path $PHP_LOCATION "ext")
    }
    $content = Get-ConfigFileContent $templatePath $configs
    Set-Content $phpIni $content
    Write-JujuLog "Generated php.ini file"
    return $phpIni
}

# Create Drupal's settings.php file. Full path to the file is returned in
# case the file is successfully generated. If settings.php file already
# exists, it will be overridden. This is intended as the settings.php file
# must be re-generated once the charm leaves a relation with a database
# charm or the memcached charm.
function Write-DrupalSettingsFile {
    $settingsPhp = Join-Path $DRUPAL_ROOT "sites\default\settings.php"
    if (Test-Path $settingsPhp) {
        # Remove read-only flag if the file already exists
        Set-ItemProperty $settingsPhp -Name IsReadOnly -Value $false
    }

    $settings = @('<?php')
    $template = Join-Path (Get-TemplatesDir) "drupal\default.settings.php"
    #
    # Add default Drupal settings
    #
    $settings += Get-Content $template

    $dbContext = Get-DatabaseContext
    if ($dbContext.Count) {
        #
        # Add database settings
        #
        $template = Join-Path (Get-TemplatesDir) "drupal\databases.settings.php"
        $configs = @{
            "db_name"     = $dbContext['db_name'];
            "db_username" = $dbContext['db_user_name'];
            "db_password" = $dbContext['db_user_password'];
            "db_host"     = $dbContext['db_host'];
            "db_driver"   = $dbContext['db_type']
        }
        $settings += Get-ConfigFileContent $template $configs
    }

    $memcachedCtx = Get-MemcachedContext
    if ($memcachedCtx.Count) {
        #
        # Add memcached settings
        #
        $template = Join-Path (Get-TemplatesDir) "drupal\memcached.settings.php"
        $configs = @{
            "memcached_host" = $memcachedCtx['cache_host'];
            "memcached_port" = $memcachedCtx['cache_tcp_port']
        }
        $settings += Get-ConfigFileContent $template $configs
    }

    Set-Content $settingsPhp $settings
    Set-ItemProperty $settingsPhp -Name IsReadOnly -Value $true
    Write-JujuLog "Generated settings file for Drupal"
}

# Installs the PHP extension for Memcache. The archive with the extension is
# downloaded from the Internet and checked for integrity. After unzipping into
# a temporary location, the extension is moved in the PHP extensions
# directory and proper permissions are set up for it.
function Install-MemcachePHPExtension {
    $extDir = Join-Path $PHP_LOCATION "ext"
    if (Test-Path "$extDir\php_memcache.dll") {
        Write-JujuLog "PHP Memcache extension already installed"
        return
    }

    Write-JujuLog "Installing Memcahe PHP extension"

    Start-ExternalCommand { iisreset.exe /STOP } -ErrorMessage "Failed to stop IIS services."
    $tempDir = Join-Path $env:TEMP "memcache-extension"
    if (Test-Path $tempDir) {
        Remove-Item -Force -Recurse $tempDir
    }
    $memcacheUrl = Get-JujuCharmConfig -Scope 'memcache-extension-url'
    $zipPath = Get-DownloadedFilePath $memcacheUrl
    Install-ZipArchive $zipPath $tempDir
    Move-Item "$tempDir\php_memcache.dll" $extDir -Force
    Remove-Item $tempDir -Recurse -Force
    Start-ExternalCommand -Command {
        # Add read and execute permissions for the extension file.
        icacls.exe "$extDir\php_memcache.dll" /grant ".\IIS_IUSRS:(RX)"
    } -ErrorMessage "Failed to set read and execute permissions."
    Start-ExternalCommand { iisreset.exe /START } -ErrorMessage "Failed to start IIS server"

    Write-JujuLog "Finished installing Memcache PHP extension."
}

# Enables memcache modules for Drupal using drush commands.
function Enable-MemcacheModule {
    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT en memcache -y
    } -ErrorMessage "Failed to enable memcache module."
    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT en memcache_admin -y
    } -ErrorMessage "Failed to enable memcache admin interface."
}

# Initializes the Drupal database using 'drush site-install' command.
# It will create the database schema and do the proper initialization.
# If the database is already initialized, this function returns early.
function Initialize-Database {
    Param(
        [Parameter(Mandatory=$true)]
        [HashTable]$DBContext
    )

    if ($DBContext['db_type'] -eq "pgsql") {
        # PostgreSQL doesn't accept plain text passwords transmitted via
        # cmd line. It uses environment variable $env:PGPASSWORD for passwords
        $env:PGPASSWORD = $DBContext['db_user_password']
    }

    # Create the database connection url using the parameters from $DBContext
    $escapedPassword = [uri]::EscapeDataString($DBContext['db_user_password'])
    $dbUrl = $DBContext['db_type'] + '://' +
             $DBContext['db_user_name'] + ':' +
             $escapedPassword + '@' +
             $DBContext['db_host'] + '/' +
             $DBContext['db_name']
    $adminEmail = Get-JujuCharmConfig -Scope 'admin-mail'
    $adminName = Get-JujuCharmConfig -Scope 'admin-user-name'
    $adminPassword = Get-JujuCharmConfig -Scope 'admin-password'
    $siteName = Get-JujuCharmConfig -Scope 'site-name'
    $siteEmail = Get-JujuCharmConfig -Scope 'site-mail'
    Write-JujuLog "Initializing the database"

    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT  site-install standard `
           --db-url=$dbUrl `
           --account-mail=$adminEmail `
           --account-name=$adminName `
           --account-pass=$adminPassword `
           --site-name=$siteName `
           --site-mail=$siteEmail `
           --y `
           install_configure_form.update_status_module='array(FALSE,FALSE)'
    } -ErrorMessage "Failed to initialize Drupal database"
}

# Installs all the MSSQL prerequisites (MSSQL native client, ODBC driver and
# MSSQL command-line utilities) and adds MSSQL bin folder to user path.
function Install-MSSQLPrerequisites {
    $packages = @(
        @{
            'url' = Get-JujuCharmConfig -Scope 'sncli-url';
            'name' = "Microsoft SQL Server 2012 Native Client ";
            'unattended_args' = @("/q", "ADDLOCAL=ALL","IACCEPTSQLNCLILICENSETERMS=YES")
        };
        @{
            'url' = Get-JujuCharmConfig -Scope 'sodbc-url';
            'name' = "Microsoft ODBC Driver 11 for SQL Server";
            'unattended_args' = @("/q", "ADDLOCAL=ALL", "IACCEPTMSODBCSQLLICENSETERMS=YES")
        };
        @{
            'url' = Get-JujuCharmConfig -Scope 'scmd-url';
            'name' = "Microsoft SQL Server 2012 Command Line Utilities ";
            'unattended_args' = @("/q", "ADDLOCAL=ALL", "IACCEPTMsSqlCmdLnUtilsLICENSETERMS=YES")
        };
    )
    foreach($pkg in $packages) {
        if (Get-ComponentIsInstalled $pkg['name']) {
            Write-JujuLog ($pkg['name'] + " is already installed")
            continue
        }
        $installerPath = Get-DownloadedFilePath $pkg['url']
        Install-Package $installerPath $pkg['unattended_args']
    }
    Add-ToUserPath (Join-Path $env:ProgramFiles "Microsoft SQL Server\110\Tools\Binn")
}

# Installs the MSSQL PHP extensions by downloading the archive from the
# Internet, extracting it into a temp directory and moving the MSSQL
# extensions into the extensions directory. If the extensions already exit
# then this function returns early.
function Install-PHPSQLServerDrivers {
    $extDir = Join-Path $PHP_LOCATION "ext"
    if ((Test-Path "$extDir\php_sqlsrv_54_nts.dll") -and
        (Test-Path "$extDir\php_pdo_sqlsrv_54_nts.dll")) {
        Write-JujuLog "MSSQL PHP drivers already installed"
        return
    }

    Write-JujuLog "Installing PHP MSSQL drivers"

    Start-ExternalCommand { iisreset.exe /STOP } -ErrorMessage "Failed to stop IIS services."
    # Extracts PHP MSSQL extensions into a temporary directory
    # and move the dll files to the extensions directory
    $tmpDir = Join-Path $env:TEMP "SQLDrivers"
    if (Test-Path $tmpDir) {
        Remove-Item -Recurse -Force $tmpDir
    }
    $url = Get-JujuCharmConfig -Scope 'php-mssql-drivers-url'
    $installerPath = Get-DownloadedFilePath $url
    Install-Package $installerPath -ArgumentList @("/Q", "/T:$tmpDir", "/C")
    # Move PHP extensions to PHP extensions directory
    $sqlsrv = Join-Path $tmpDir "php_sqlsrv_54_nts.dll"
    $sqlsrvPDO = Join-Path $tmpDir "php_pdo_sqlsrv_54_nts.dll"
    Move-Item $sqlsrv $extDir -Force
    Move-Item $sqlsrvPDO $extDir -Force
    # Fix permissions for the MSSQL PHP extensions
    Start-ExternalCommand {
        icacls.exe "$extDir\php_sqlsrv_54_nts.dll" /grant ".\IIS_IUSRS:(RX)"
    } -ErrorMessage "Failed to set read and execute permissions."
    Start-ExternalCommand {
        icacls.exe "$extDir\php_pdo_sqlsrv_54_nts.dll" /grant ".\IIS_IUSRS:(RX)"
    } -ErrorMessage "Failed to set read and execute permissions."
    Start-ExternalCommand { iisreset.exe /START } -ErrorMessage "Failed to start IIS services."
    Remove-Item $tmpDir -Recurse -Force

    Write-JujuLog "Finished installing PHP MSSQL drivers."
}

function New-DrupalFirewallRule {
    $firewallRuleName = 'Drupal-IIS-Juju-Charm'
    $rule = Get-NetFirewallRule | Where-Object { $_.Name -eq $firewallRuleName }
    if ($rule) {
        Remove-NetFirewallRule -Name $firewallRuleName
    }
    $drupalPort = Get-JujuCharmConfig -Scope 'iis-website-port'
    New-NetFirewallRule -Name $firewallRuleName `
                        -DisplayName $firewallRuleName `
                        -Enabled True -Direction Inbound `
                        -LocalPort $drupalPort -Protocol TCP
}

function Set-PeerSettings {
    $settings = @{
        'db-connected' = Get-CharmState -Key 'DatabaseIsConnected';
        'memcache-installed' = Get-CharmState -Key 'MemcacheIsInstalled'
    }
    $rids = Get-JujuRelationIds 'cluster'
    foreach ($r in $rids) {
        try {
            Set-JujuRelation -Settings $settings -RelationId $r
        } catch {
            Write-JujuError "Failed to set drupal-iis cluster relation settings."
        }
    }
}

# Installs IIS, PHP and Drupal core files. IIS is configured with PHP and
# Drupal website is created.
function Start-InstallHook {
    # Set machine to use high performance settings.
    try {
        Set-PowerProfile -PowerProfile Performance
    } catch {
        # No need to error out the hook if this fails.
        Write-JujuWarning "Failed to set power scheme."
    }
    Start-TimeResync
    Install-IIS
    Install-DrupalPrerequisites

    Write-JujuLog "Installing PHP"
    if (!(Test-Path $PHP_LOCATION)) {
        $url = Get-JujuCharmConfig -Scope 'php-url'
        $zipPath = Get-DownloadedFilePath $url
        Install-ZipArchive $zipPath $PHP_LOCATION
    } else {
        Write-JujuLog "PHP archive already extracted"
    }
    Write-PHPIniFile
    Add-ToUserPath $PHP_LOCATION
    Write-JujuLog "Finished installing PHP"

    # Install Drupal 7 core files, create Drupal temporary directory
    # and set proper permissions for the IIS user.
    Write-JujuLog "Installing Drupal"
    if (!(Test-Path $DRUPAL_ROOT)) {
        $url = Get-JujuCharmConfig -Scope 'drupal-url'
        $zipPath = Get-DownloadedFilePath $url
        Install-ZipArchive $zipPath $DRUPAL_ROOT -WithFolder
    } else {
        Write-JujuLog "Drupal archive already extracted"
    }
    $drupalTmp = Join-Path $DRUPAL_ROOT "sites\default\files\tmp"
    if (!(Test-Path -Path $drupalTmp)) {
        New-Item -ItemType Directory -Path $drupalTmp
    }
    Start-ExternalCommand {
        icacls.exe "$DRUPAL_ROOT\*" /grant "IUSR:(OI)(CI)(F)" /t /q
    } -ErrorMessage "Failed to set permissions to IIS user for '$DRUPAL_ROOT' folder"
    Write-JujuLog "Finished installing Drupal"

    # Configure IIS and create Drupal website pointing to its core files
    # downloaded before.
    Start-ConfigureIIS
    Import-Module WebAdministration
    $drupalPort = Get-JujuCharmConfig -Scope 'iis-website-port'
    if (!(Get-Website -Name $DRUPAL_WEBSITE_NAME)) {
        New-WebSite -Name $DRUPAL_WEBSITE_NAME -Port $drupalPort -PhysicalPath $DRUPAL_ROOT
    }
    # Open firewall port and Juju for Drupal website
    New-DrupalFirewallRule
    Open-JujuPort "$drupalPort/tcp"
    Set-CharmState -Key 'IISWebsitePort' -Value $drupalPort

    $ctxt = Get-ClusterContext
    if ($ctxt.Count) {
        Set-CharmState -Key 'DatabaseIsConnected' -Value $ctxt['db-connected']
        Set-CharmState -Key 'MemcacheIsInstalled' -Value $ctxt['memcache-installed']
    } else {
        Set-CharmState -Key 'DatabaseIsConnected' -Value 'False'
        Set-CharmState -Key 'MemcacheIsInstalled' -Value 'False'
    }
}

# At the end of the execution of config-changed hook, the following configs
# are changed: 'admin-password', 'site-mail' and 'site-name'. They will have
# the values given in config.yaml. If the database is not yet initialized or
# the unit is not master, this hook will be skipped.
function Start-ConfigChangedHook {
    Import-Module WebAdministration
    $newPort = Get-JujuCharmConfig -Scope 'iis-website-port'
    $oldPort = Get-CharmState -Key 'IISWebsitePort'
    if ($newPort -ne $oldPort) {
        $bindingInfo = (Get-WebBinding -Name $DRUPAL_WEBSITE_NAME).bindingInformation
        Set-WebBinding -Name $DRUPAL_WEBSITE_NAME -PropertyName Port `
                       -Value $newPort -BindingInformation $bindingInfo
        New-DrupalFirewallRule
        Close-JujuPort "$oldPort/tcp"
        Open-JujuPort "$newPort/tcp"
        Set-CharmState -Key 'IISWebsitePort' -Value $newPort
        Start-WebsiteRelationJoinedHook
    }

    if (!(Confirm-Leader)) {
        Write-JujuLog ("Unit is not leader. Skipping config-changed hook as " +
                       "it only makes changes to database.")
        return
    }
    if (!(Get-DatabaseIsConnected)) {
        Write-JujuLog "Database is not connected. Skipping the rest of config-changed hook"
        return
    }

    $env:PATH = Get-UserPath
    $adminPassword = Get-JujuCharmConfig -Scope 'admin-password'
    $adminUser = Get-JujuCharmConfig -Scope 'admin-user-name'
    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT upwd --password=$adminPassword $adminUser
    } -ErrorMessage "Failed to set admin user password."

    $siteEmail = Get-JujuCharmConfig -Scope 'site-mail'
    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT variable-set site_mail "$siteEmail"
    } -ErrorMessage "Failed to set site e-mail."

    $siteName = Get-JujuCharmConfig -Scope 'site-name'
    Start-ExternalCommand {
        drush -r $DRUPAL_ROOT variable-set site_name "$siteName"
    } -ErrorMessage "Failed to set site name."
}

function Start-StartHook {
    Import-Module WebAdministration
    Start-Website $DRUPAL_WEBSITE_NAME
}

function Start-StopHook {
    Import-Module WebAdministration
    Stop-Website $DRUPAL_WEBSITE_NAME
}

# Installs the prerequisites needed (such as MySQL client) for the
# database relation.
function Start-MySQLRelationJoinedHook {
    $env:PATH = Get-UserPath
    if ((Confirm-JujuRelationCreated -Relation 'mssql-db') -or
        (Confirm-JujuRelationCreated -Relation 'pgsql-db')) {
        Throw ("Cannot create MySQL database relation. " +
               "Another db-relation already created. Please resolve the juju unit, " +
               "remove both database relations and add the desired one.")
    }
    $settings = @{
        'username' = Get-JujuCharmConfig -Scope 'database-user-name';
        'database' = Get-JujuCharmConfig -Scope 'database-name';
        'hostname' = Get-JujuUnit -Attribute 'private-address'
    }
    $rids = Get-JujuRelationIds 'mysql-db'
    foreach ($r in $rids) {
        try {
            Set-JujuRelation -Settings $settings -RelationId $r
        } catch {
            Write-JujuError "Failed to set MySQL relation settings."
        }
    }
    $pkgName = "MySQL Server 5.5"
    if (!(Get-ComponentIsInstalled $pkgName)) {
        $url = Get-JujuCharmConfig -Scope 'mysql-url'
        $installerPath = Get-DownloadedFilePath $url
        Install-Package $installerPath -ArgumentList @('/q')
    } else {
        Write-JujuLog ("$pkgName is already installed")
    }
    Add-ToUserPath (Join-Path ${env:ProgramFiles(x86)} 'MySQL\MySQL Server 5.5\bin')
}

# Sets the relation parameters for the pgsql-db relation and installs
# the prerequisites (Windows PostgreSQL client and VC12).
function Start-PgSQLRelationJoinedHook {
    $env:PATH = Get-UserPath
    if ((Confirm-JujuRelationCreated -Relation 'mssql-db') -or
        (Confirm-JujuRelationCreated -Relation 'mysql-db')) {
        Throw ("Cannot create MySQL database relation. " +
               "Another db-relation already created. Please resolve the juju unit, " +
               "remove both database relations and add the desired one.")
    }

    $settings = @{
        'roles' = 'postgres';
        'database' = Get-JujuCharmConfig -Scope 'database-name'
    }
    $rids = Get-JujuRelationIds 'pgsql-db'
    foreach ($r in $rids) {
        try {
            Set-JujuRelation -Settings $settings -RelationId $r
        } catch {
            Write-JujuError "Failed to set PGSQL relation settings."
        }
    }

    $pkgName = "Microsoft Visual C++ 2013 x86 Additional Runtime - 12.0.21005"
    if (!(Get-ComponentIsInstalled $pkgName)) {
        Write-JujuLog ("$pkgName is already installed")
    } else {
        $url = Get-JujuCharmConfig -Scope 'vc12-url'
        $installerPath = Get-DownloadedFilePath $url
        Install-Package $installerPath -ArgumentList @('/q')
    }

    $pgsqlLocation = Join-Path $env:SystemDrive "pgsql"
    if (!(Test-Path $pgsqlLocation)) {
        $url = Get-JujuCharmConfig -Scope 'pgsql-url'
        $zipPath = Get-DownloadedFilePath $url
        Install-ZipArchive $zipPath $pgsqlLocation -WithFolder
    } else {
        Write-JujuLog "PGSQL archive already extracted"
    }
    Add-ToUserPath (Join-Path $pgsqlLocation "bin")
}

# Set the relation parameter (database name) and installs all the MSSQL
# prerequisites (such as: PHP MSSQL Drupal drivers, MSSQL Native client,
# MSSQL command-line utilities, etc.).
function Start-MSSQLRelationJoinedHook {
    $env:PATH = Get-UserPath
    if ((Confirm-JujuRelationCreated -Relation 'mysql-db') -or
        (Confirm-JujuRelationCreated -Relation 'pgsql-db')) {
        Throw ("Cannot create MySQL database relation. " +
               "Another db-relation already created. Please resolve the juju unit, " +
               "remove both database relations and add the desired one.")
    }

    $settings = @{
        'user' = Get-JujuCharmConfig -Scope 'database-user-name';
        'database' = Get-JujuCharmConfig -Scope 'database-name'
    }
    $rids = Get-JujuRelationIds 'mssql-db'
    foreach ($r in $rids) {
        try {
            Set-JujuRelation -Settings $settings -RelationId $r
        } catch {
            Write-JujuError "Failed to set MSSQL relation settings."
        }
    }
    Install-MSSQLPrerequisites
    Install-PHPSQLServerDrivers
    # Download and install MSSQL Drupal drivers
    $url = Get-JujuCharmConfig -Scope 'mssql-drivers-url'
    $zipPath = Get-DownloadedFilePath $url
    $sqlsrvModule = Join-Path $DRUPAL_ROOT "sites\all\modules\sqlsrv"
    Install-ZipArchive $zipPath $sqlsrvModule -WithFolder
    Start-ExternalCommand {
        icacls.exe "$sqlsrvModule\*" /grant "IUSR:(OI)(CI)(F)" /t /q
    } -ErrorMessage "Failed to set permissions for '$sqlsrvModule'"
    $includeDb = Join-Path $DRUPAL_ROOT "includes\database"
    Copy-Item -Recurse -Force "$sqlsrvModule\sqlsrv" $includeDb
}

function Start-DBRelationChanged {
    $env:PATH = Get-UserPath
    if (Confirm-Leader) {
        if (!(Get-DatabaseIsConnected)) {
            $DBContext = Get-DatabaseContext
            if (!$DBContext.Count) {
                Write-JujuLog "Database context not ready. Skipping"
                return
            }
            Initialize-Database $DBContext
            $drupalTmp = Join-Path $DRUPAL_ROOT "sites\default\files\tmp"
            Start-ExternalCommand -Command {
                drush -r $DRUPAL_ROOT vset file_temporary_path $drupalTmp
            } -ErrorMessage "Failed to set Drupal temporary directory"
            Set-CharmState -Key 'DatabaseIsConnected' -Value 'True'

            Write-DrupalSettingsFile
            $memcachedCtx = Get-MemcachedContext
            if ((Get-MemcacheIsInstalled) -and $memcachedCtx.Count) {
                Enable-MemcacheModule
            }
            Set-PeerSettings
        }
    }
    Write-DrupalSettingsFile
}

# Re-generates settings.php and deletes Drupal MSSQL modules files
function Start-MSSQLRelationBrokenHook {
    $sqlsrvModule = Join-Path $DRUPAL_ROOT "sites\all\modules\sqlsrv"
    if (Test-Path $sqlsrvModule) {
        Remove-Item $sqlsrvModule -Recurse -Force
    }
    $sqlsrvIncludeDB = Join-Path $DRUPAL_ROOT "includes\database\sqlsrv"
    if (Test-Path $sqlsrvIncludeDB) {
        Remove-Item $sqlsrvIncludeDB -Recurse -Force
    }
}

function Start-DBRelationBroken {
    Write-DrupalSettingsFile
    if (Confirm-Leader) {
        Set-CharmState -Key 'DatabaseIsConnected' -Value 'False'
        Set-PeerSettings
    }
}

# Installs and enables the memcache module. Zip archive with the module is
# downloaded from the Internet, unzipped in the Drupal modules directory and
# configurations are added to settings.php to enable the module.
function Start-CacheRelationChangedHook {
    $env:PATH = Get-UserPath
    $memcachedCtx = Get-MemcachedContext
    if (!$memcachedCtx.Count) {
        Write-JujuLog "Memcached context not ready. Skipping"
        return
    }
    Write-JujuLog "Installing Drupal Memcache module"
    Install-MemcachePHPExtension
    $memcacheModuleDir = Join-Path $DRUPAL_ROOT "sites\all\modules\memcache"
    try {
        $url = Get-JujuCharmConfig -Scope 'memcache-url'
        $zipPath = Get-DownloadedFilePath $url
        Install-ZipArchive $zipPath $memcacheModuleDir -WithFolder
        # Fix permissions for IIS_IUSRS
        Start-ExternalCommand {
            icacls.exe "$memcacheModuleDir\*" /grant "IUSR:(OI)(CI)(F)" /t /q
        } -ErrorMessage "Failed to set permissions for '$memcacheModuleDir'"
    } catch {
        # In case extraction failed at some point, remove the files resulted
        # from the uncompleted unzipping.
        Remove-Item -Recurse -Force $memcacheModuleDir
        Throw $_
    }
    Write-DrupalSettingsFile
    if ((Confirm-Leader) -and (Get-DatabaseIsConnected)) {
        # Enable memcache modules. This is done only on the master unit as
        # it requires interaction with the database. On the slave units
        # only configuration to the memcached server is made.
        Enable-MemcacheModule
    }
    Write-DrupalSettingsFile
    Set-CharmState -Key 'MemcacheIsInstalled' -Value 'True'
    Set-PeerSettings
    Write-JujuLog "Finished installing Memcache module."
}

# Disables and removes the memcache modules
function Start-CacheRelationBrokenHook {
    $env:PATH = Get-UserPath
    if ((Confirm-Leader) -and (Get-DatabaseIsConnected)) {
        try {
            # Disable the Memcache module
            Start-ExternalCommand {
                drush -r $DRUPAL_ROOT dis memcache -y
            } -ErrorMessage "Failed to disable memcache module."
            Start-ExternalCommand {
                drush -r $DRUPAL_ROOT dis memcache_admin -y
            } -ErrorMessage "Failed to disable memcache admin interface."
        } catch {
            Write-JujuLog ("Cannot disable Memcache module. Probably the " +
                           "database relation was dropped.")
        }
    }
    # Remove the modules files
    $memcacheModule = Join-Path $DRUPAL_ROOT "sites\all\modules\memcache"
    if (Test-Path $memcacheModule) {
        Remove-Item -Recurse -Force $memcacheModule
    }
    Set-CharmState -Key 'MemcacheIsInstalled' -Value 'False'
    Set-PeerSettings
    # Re-generate settings.php without memcached configurations
    Write-DrupalSettingsFile
}

function Start-WebsiteRelationJoinedHook {
    $settings = @{
        "hostname" = Get-JujuUnit -Attribute 'private-address';
        "port"     = Get-JujuCharmConfig -Scope 'iis-website-port';
    }
    $rids = Get-JujuRelationIds 'website'
    foreach ($r in $rids) {
        try {
            Set-JujuRelation -Settings $settings -RelationId $r
        } catch {
            Write-JujuError "Failed to set website relation settings."
        }
    }
}

function Start-ClusterRelationChangedHook {
    $dbConnected = Get-JujuRelation -Attribute 'db-connected'
    Set-CharmState -Key 'DatabaseIsConnected' -Value $dbConnected
    $memcacheInstalled = Get-JujuRelation -Attribute 'memcache-installed'
    Set-CharmState -Key 'MemcacheIsInstalled' -Value $memcacheInstalled
}
